import datetime
import json
import os
import pytz

import quicklogging
import yaml

from . import utils


class CheckItem(yaml.YAMLObject):
    yaml_tag = "!CheckItem"

    def __init__(self, data):
        self.label = data["label"]
        self.value = data["value"]
        self.extra = data["extra"] or None

    @classmethod
    def from_yaml(cls, loader, node):
        return cls(loader.construct_mapping(node))

    @classmethod
    def to_yaml(cls, dumper, liveobject):
        return dumper.represent_mapping(
            cls.yaml_tag,
            {
                "label": liveobject.label,
                "value": liveobject.value,
                "extra": liveobject.extra,
            }
        )


class Check(yaml.YAMLObject):
    yaml_tag = "!Check"

    def __init__(self, data, time):
        # {'group': '2p', 'host': 'magellan.2p', 'graph_category': 'disk',
        # 'graph_title': 'Disk usage in percent', 'warning': [{'label':
        # '/mnt/backups', 'value': '93.04', 'w': ':92', 'c': ':98', 'extra':
        # ''}]}

        self.time = time
        self.host = data["host"]
        self.group = data["group"]

        # first get works with json, second with yaml
        self.category = data.get("graph_category") or data["category"]
        self.title = data.get("graph_title") or data["title"]

        if "level" in data:
            # from yaml cache
            self.level = data["level"]
        # other possibilities: from http json push
        elif "error" in data:
            self.level = "error"
        elif "warning" in data:
            self.level = "warning"
        elif "unknown" in data:
            self.level = "unknown"
        else:
            self.level = "ok"

        # data["subchecks"] exists when loading yaml cache
        self.subchecks = data.get("subchecks") or []

        # fill from munin structure
        for level in ("error", "warning", "unknown", "ok"):
            for item in data.get(level, ()):
                self.subchecks.append(CheckItem(item))

        self.key = f"{self.group} {self.host} {self.category} {self.title}"

    @property
    def icon(self):
        if self.category == "disk":
            return "harddisk.png"
        return "point.png"

    def diff(self, other):
        if other is None:
            return f"nothing → {self.time} - {self.level}"
        if self.level == other.level:
            return None
        return f"{other.time} - {other.level} → { self.time } - {self.level}"

    @classmethod
    def to_yaml(cls, dumper, liveobject):
        return dumper.represent_mapping(
            cls.yaml_tag,
            {
                "category": liveobject.category,
                "group": liveobject.group,
                "host": liveobject.host,
                "level": liveobject.level,
                "time": liveobject.time,
                "title": liveobject.title,
                "subchecks": liveobject.subchecks,
            }
        )

    @classmethod
    def from_yaml(cls, loader, node):
        mapping = loader.construct_mapping(node)
        return Check(mapping, mapping["time"])


class CheckStore(yaml.YAMLObject):
    yaml_tag = "!CheckStore"

    def __init__(self, restored_backend=None):
        self._backend = {}
        self._decoder = json.JSONDecoder()

        if restored_backend:
            for check in restored_backend:
                self._insert_check(check)

    def consume(self, jsontext):
        """Register checks from a concatenated json stream"""
        now_utc = datetime.datetime.now(pytz.timezone("UTC"))
        text = jsontext.lstrip()
        while text:
            jsonobj, index = self._decoder.raw_decode(text)

            self._consume_one(jsonobj, now_utc)

            # truncate consumed json (left part), then leading spaces.
            text = text[index:].lstrip()

    def _consume_one(self, data, now_utc):
        """Makes a check from a deserialized single json check representation

        :param dict data:
        :param datetime.datetime now_utc:
        """
        utils.purge_none(data)

        check = Check(data, now_utc)
        self._insert_check(check)

    def _insert_check(self, check):
        check_key = check.key

        previous_state = self._backend.get(check_key)
        diff = check.diff(previous_state)
        if diff:
            quicklogging.info("%s: %s", check_key, diff)

        self._backend[check_key] = check

    def getchecks(self):
        return self._backend.values()

    @classmethod
    def from_yaml(cls, loader, node):
        return CheckStore(loader.construct_sequence(node))

    @classmethod
    def to_yaml(cls, dumper, liveobject):
        return dumper.represent_sequence(
            cls.yaml_tag,
            liveobject._backend.values()
        )

    @classmethod
    def restore(cls, savefile):
        with open(savefile, "r") as savefd:
            cached_checks = yaml.load(savefd, yaml.FullLoader).values()
        return CheckStore(restored_backend=cached_checks)

    def save(self, savefile):
        """Dumps current data into a yaml file

        First: writes yaml to swp file
        Second moves this file in place

        As to avoid inconsistency in the yaml file if anything happens while
        saving.
        """
        savefile_swp = "{}.swp".format(savefile)
        with open(savefile_swp, 'w') as savefd:
            yaml.dump(self._backend, savefd)
        os.rename(savefile_swp, savefile)



"""munin2widget

HTTP server converting munin-json to html reports
that you can display on your smartphone (html widget)

"""

APP_NAME = 'munin2smartphone'

APP_DESCRIPTION = (
    "HTTP server converting munin-json to html reports "
    "that you can display on your smartphone (html widget)"
)


__all__ = ("APP_NAME", "APP_DESCRIPTION")



import copy
import datetime
import jinja2
import json
import logging
import os
import pytz

import quicklogging

from . import APP_NAME
from . import datastore
from . import utils

DATE_FORMAT = "%H:%M:%S %d/%m/%Y %Z%z"


class Report:
    def __init__(
            self,
            outputdir,
            name,
            template_loader,
            accepted_checklevels=None,
            ignored_keyvalues=None,
    ):
        """
        :param sequence accepted_checklevels: if None, all levels are accepted
            list containing "warning", "critical", "unknown", "ok"
        :param sequence ignored_keyvalues: if None, all checks are accepted
        """
        self.name = name
        self.accepted_checklevels = accepted_checklevels
        self.ignored_keyvalues = ignored_keyvalues
        outputfile = os.path.join(outputdir, name)
        self.filename = f"{outputfile}.html"
        self.filename_swp = f"{self.filename}.swp"

        # set in config
        templateEnv = jinja2.Environment(loader=template_loader)
        TEMPLATE_FILE = "report.html.jinja2"
        self.template = templateEnv.get_template(TEMPLATE_FILE)

    def _filter_checklevels(self, all_checks):
        for check in all_checks:
            if self.accepted_checklevels is None:
                yield check
            elif check.level in self.accepted_checklevels:
                yield check

    def _filter_ignored(self, all_checks):
        if self.ignored_keyvalues is None:
            for check in all_checks:
                yield check
            return
        for attr, value in self.ignored_keyvalues.items():
            for check in all_checks:
                if getattr(check, attr) == value:
                    continue
                yield check

    def _all_filters(self, all_checks):
        residual = copy.copy(all_checks)
        for filterfunc in (
                self._filter_checklevels,
                self._filter_ignored,
        ):
            residual = list(filterfunc(residual))

        return residual

    def generate(self, all_checks, printable_date):
        displayed = list(self._all_filters(all_checks))

        with open(self.filename_swp, "w") as reportfd:
            reportfd.write(
                self.template.render(
                    supervision_data=displayed,
                    date=printable_date,
                )
            )
        os.rename(self.filename_swp, self.filename)

        quicklogging.debug(
            "Updated: %s: %s with %d records",
            printable_date,
            self.filename,
            len(displayed)
        )


class Munin2Widget:
    def __init__(self, config, restore=False):
        self.cache_file = os.path.join(config.cache_directory, "cache.yaml")
        self._reports = list(
            self._gen_reports(config.outputdir, config.reports)
        )
        checkstore = self._get_store(restore)
        assert isinstance(checkstore, datastore.CheckStore)
        self.store = checkstore
        self._tz_utc = pytz.timezone("UTC")
        self._tz_user = config.timezone

    def consume(self, text):
        """Handles request content

        :param str text:
        """
        self.store.consume(text)
        self.gen_html()
        self._update_cache()

    def _gen_reports(self, outputdir, reportcfg_dict):
        template_loader = jinja2.PackageLoader(APP_NAME, "templates")
        for name, reportcfg in reportcfg_dict.items():
            yield Report(
                outputdir,
                name,
                template_loader,
                **reportcfg
            )

    def _get_store(self, restore):
        if restore:
            try:
                return datastore.CheckStore.restore(self.cache_file)
            except OSError:
                logging.exception(
                    "Error opening cache file %s - this is not fatal, "
                    "starting with no cache",
                    self.cache_file
                )
        return datastore.CheckStore()

    def _update_cache(self):
        self.store.save(self.cache_file)

    def gen_html(self):
        now_utc = datetime.datetime.now(self._tz_utc)
        now_local = now_utc.astimezone(self._tz_user)
        printable_date = now_local.strftime(DATE_FORMAT)
        data = list(self.store.getchecks())
        for report in self._reports:
            report.generate(data, printable_date)

    def filter_essential(self, obj):
        if obj.level not in self.checklevels:
            return False
        for attr, value in self.ignored.items():
            if getattr(obj, attr) == value:
                return False
        return True

    def json_load_concatenated(self, text):
        text = text.lstrip()
        decoder = json.JSONDecoder()
        while text:
            obj, index = decoder.raw_decode(text)
            utils.purge_none(obj)

            if self.should_display(obj):
                yield obj

            text = text[index:].lstrip()



import aiohttp.web

import quicklogging

from . import widget


class Server:
    def __init__(self, config, restore=False):
        self.config = config
        self.processor = widget.Munin2Widget(
            self.config,
            restore=restore
        )

    def run(self):
        app = aiohttp.web.Application()

        app.router.add_route("*", "/{tail:.*}", self.handle_request)

        quicklogging.info("all set up")

        aiohttp.web.run_app(
            app,
            port=self.config.port,
            host=self.config.listening_address,
        )

    async def handle_request(self, request):
        if request.method == "GET":
            quicklogging.debug("%s", request)
        if request.method == "POST":
            text = await request.text()
            self.processor.consume(text)

        return aiohttp.web.Response(text="Hello, world")



import argparse
import errno
import logging
import os
import pytz

try:
    import importlib.metadata as importlib_metadata
except ImportError:
    import importlib_metadata

import coloredlogs
import quicklogging
import xdg.BaseDirectory

from . import APP_NAME
from . import APP_DESCRIPTION
from .exceptions import ConfigError


class ConfigValueUnspecified:
    """This special object works for incremental values"""
    def __add__(self, value):
        return value


class Config:
    _UNSPECIFIED = ConfigValueUnspecified()
    DEFAULT_TIMEZONE = "Europe/Paris"
    DEFAULT_PORT = 8765
    DEFAULT_LISTENING_ADDRESS = "127.0.0.1"
    DEFAULT_REPORTS = {
        "report": {
            "accepted_checklevels": ("warning", "critical"),
            "ignored_keyvalues": {"title": "Pending packages"},
        },
        "all": {},
    }

    def __init__(self, withcli=True, config=None):
        self._cfgbackend = self._config_defaults()

        if config:
            self._cfgbackend.update(config)

        if withcli:
            cli_config = vars(Config.cli_config())
            for key, value in cli_config.items():
                if value is Config._UNSPECIFIED:
                    continue
                self._cfgbackend[key] = value

        self._sanity()

    def _sanity(self):
        self._sanity_loglevel()
        self._sanity_timezone()
        self._sanity_cachedir()

    def _sanity_timezone(self):
        timezone = self._cfgbackend["timezone"]
        tz_object = pytz.timezone(timezone)
        self._cfgbackend["timezone"] = tz_object
        quicklogging.debug("Times displayed in timezone: %s", timezone)

    def _sanity_loglevel(self):
        verbosity = self._cfgbackend["verbose"]
        logger = quicklogging.get_logger()
        if verbosity == 0:
            level, logmsg = logging.ERROR, "error"
        elif verbosity == 1:
            level, logmsg = logging.WARNING, "warning"
        elif verbosity == 2:
            level, logmsg = logging.INFO, "info"
        else:
            level, logmsg = logging.DEBUG, "debug"

        if self._cfgbackend["coloredlogs"]:
            coloredlogs.install(level=level)
        else:
            logging.basicConfig(level=level)

        logger.info("Logging level set to %s", logmsg)

    def _sanity_cachedir(self):
        cache_directory = self.cache_directory
        quicklogging.debug("Cache directory is %s", cache_directory)
        if os.path.exists(cache_directory):
            if os.path.isdir(cache_directory):
                return
            raise ConfigError(
                "Cache directory ({}) exists and is not a directory.".format(
                    cache_directory
                )
            )

        # only for non default cache dir (otherwise xdg takes care of this)
        try:
            quicklogging.warning(
                "Creating cache directory %s",
                cache_directory,
            )
            os.makedirs(cache_directory)
        except OSError as err:
            raise ConfigError(
                "Error creating cache directory ({}): {}".format(
                    cache_directory,
                    errno.errorcode(err.errno)
                )
            )

    @property
    def timezone(self):
        return self._cfgbackend["timezone"]

    @property
    def port(self):
        return self._cfgbackend["port"]

    @property
    def listening_address(self):
        return self._cfgbackend["listening-address"]

    @property
    def cache_directory(self):
        return self._cfgbackend["cache_directory"]

    @property
    def outputdir(self):
        return self._cfgbackend["outputdir"]

    @property
    def reports(self):
        # TODO: read a config value for reports
        return self.DEFAULT_REPORTS

    @staticmethod
    def _config_defaults():
        default_cache_dir = xdg.BaseDirectory.save_cache_path(APP_NAME)
        default_outputdir = os.path.abspath(os.getcwd())
        return {
            "cache_directory": default_cache_dir,
            "outputdir": default_outputdir,
            "port": Config.DEFAULT_PORT,
            "listening-address": Config.DEFAULT_LISTENING_ADDRESS,
            "timezone": Config.DEFAULT_TIMEZONE,
            "verbose": 0,
            "coloredlogs": True,
        }

    @staticmethod
    def cli_config():

        confdef = Config._config_defaults()
        parser = argparse.ArgumentParser(
            description=APP_DESCRIPTION,
        )
        progversion = importlib_metadata.version("munin2smartphone")

        parser.add_argument(
            "-V",
            "--version",
            action="version",
            version="%(prog)s {version}".format(version=progversion),
        )

        parser.add_argument(
            "-v",
            "--verbose",
            help="increase output verbosity -can be called multiple times: "
            "Levels: 0 time: error, 1 time: warning, 2 times: info, 3 times:"
            "debug",
            action="count",
            default=Config._UNSPECIFIED,
        )

        parser.add_argument(
            "--coloredlogs",
            help="terminal logs are colored (using coloredlogs): default",
            default=Config._UNSPECIFIED,
            action="store_true",
        )

        parser.add_argument(
            "--no-coloredlogs",
            help="terminal logs are NOT colored: default \
            is colored logs ",
            dest="coloredlogs",
            action="store_false",
        )

        parser.add_argument(
            "-o",
            "--outputdir",
            help=f"html output directory (default: {confdef['outputdir']})",
            default=Config._UNSPECIFIED,
        )

        parser.add_argument(
            "--configfile",
            type=argparse.FileType('r', encoding='utf-8'),
            default=None,
        )

        parser.add_argument(
            "--logfile",
            type=argparse.FileType('a', encoding='utf-8'),
            default=None,
        )

        parser.add_argument(
            "--cache_directory",
            default=Config._UNSPECIFIED,
            help=f"{APP_NAME} cache directory \
            (defaults to {confdef['cache_directory']})"
        )

        parser.add_argument(
            "--port",
            type=int,
            default=Config._UNSPECIFIED,
            help=f"listening TCP port (defaults to {confdef['port']})"
        )

        parser.add_argument(
            "--listening-address",
            default=Config._UNSPECIFIED,
            help=f"listening IP address (defaults to \
            {confdef['listening-address']})",
        )

        parser.add_argument(
            "--timezone",
            default=Config._UNSPECIFIED,
            help=f"Timezone for html output -internal dates are UTC \
            (defaults to {confdef['timezone']})",
        )

        return parser.parse_args()



def purge_none(data):
    """
    Delete keys with the value ``None`` in a dictionary, recursively.

    fixed a stackoverflow algo
    """
    for key, value in list(data.items()):
        if isinstance(value, dict):
            purge_none(value)
        if value:
            continue
        else:
            del data[key]



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['munin2smartphone']

package_data = \
{'': ['*'], 'munin2smartphone': ['templates/*']}

install_requires = \
['PyYAML==5.3.1',
 'aiohttp==3.6.2',
 'coloredlogs>=14.0,<15.0',
 'jinja2>=2.11.2,<3.0.0',
 'pytz>=2020.1,<2021.0',
 'pyxdg>=0.26,<0.27',
 'quicklogging>=0.3,<0.4',
 'toml>=0.10.1,<0.11.0']

entry_points = \
{'console_scripts': ['munin2smartphone = munin2smartphone.entrypoints:run_from_cli']}

setup_kwargs = {
    'name': 'munin2smartphone',
    'version': '0.0.4',
    'description': 'Demilitarized HTTP server eating munin monitoring data and producing static pages that can be displayed securely (ie. without javascript), on your smartphone.',
    'long_description': '================\nmunin2smartphone\n================\n\nDemilitarized HTTP server eating monitoring data and producing static pages that can be displayed securely (ie. without javascript), on your smartphone.\n\n.. |made-with-python| image:: https://img.shields.io/badge/Made%20with-Python-1f425f.svg\n   :target: https://www.python.org/\n\n.. |license-GPL-2| image::  https://img.shields.io/badge/license-GPL%202-informational\n   :target: https://framagit.org/feth/munin2smartphone/-/blob/master/LICENSE_GPL_2.txt\n\n.. |license-CeCILL-2.1| image::  https://img.shields.io/badge/license-CeCILL--2.1-informational\n   :target: https://framagit.org/feth/munin2smartphone/-/blob/LICENSE_CeCILL_2.1.txt\n\n.. |project-url| image:: https://img.shields.io/badge/homepage-framagit-blue\n   :target: https://framagit.org/feth/munin2smartphone\n\n.. |repository-url| image:: https://img.shields.io/badge/repository-git%2Bhttps-blue\n   :target: https://framagit.org/feth/munin2smartphone.git\n\n.. |documentation| image:: https://readthedocs.org/projects/munin2smartphone/badge/?version=latest   :alt: Read the Docs\n\n|made-with-python| |license-GPL-2| |license-CeCILL-2.1| |project-url| |repository-url|\n\n.. figure:: docs/images/screenshot_android.png\n\n   Screen capture of *munin2smartphone* on an android phone.\n\nInstallation, usage and development documentation: |documentation|\n\ndocumentation = "https://munin2smartphone.readthedocs.io/"\n\n',
    'author': 'Feth AREZKI',
    'author_email': 'feth@majerti.fr',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://framagit.org/feth/munin2smartphone',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'entry_points': entry_points,
    'python_requires': '>=3.6',
}


setup(**setup_kwargs)



import munin2smartphone.config as config
import munin2smartphone.server as server


def run_from_cli():
    configobj = config.Config(withcli=True)
    serverobj = server.Server(configobj, restore=True)
    serverobj.run()


